/*
 *   File:  Connection.java
 *
 *   Version 1.02 v1a
 *   Copyright 1998, 1999 by Hush Communications Corporation, BWI
 */

package hushcode;

import java.applet.*;
import java.net.*;
import java.io.*;
import java.awt.*;
import java.util.*;
import hushcode.HushSHA1;
import hushcode.BlowfishCipher;
import hushcode.HushApplet;
import hushcode.Conversions;
import hushcode.ElGamalCipher;
import hushcode.HushRandom;
import hushcode.ByteQueue;


class Connection
{ 

/************************************************************
 *  The class variables and the methods that return them 
 ************************************************************/

/**
 * The following variables and methods deal with the user's folderlist.  
 * The folderlist is obtained by calling the method getFolders().
 */

  /* The number of folders used by the user
   */
  private int numberOfFolders;
  int numberOfFolders() {return numberOfFolders;}  
  
  /* An array containing the names of the user's folders
   * in the order they were received from the server
   */
  private String[] folderNameArray;
  String folderName(int x) {return folderNameArray[x];}

  /* An array containing the numbers of messages in each folder
   * in the order they were received from the server
   */
  private int[] numberOfMessagesArray;
  int numberOfMessages(int x) 
    {return numberOfMessagesArray[x];}

  /* An array containing the numbers of unread messages in each folder
   * in the order they were received from the server 
   */
  private int[] numberOfUnreadMessagesArray;
  int numberOfUnreadMessages(int x) 
    {return numberOfUnreadMessagesArray[x];}

/**
 * The following variables and methods deal with the currently open 
 * folder.  A folder is opened by calling the method openFolder(String foldername)
 */

  /* The name of the currently open folder
   */
  private String folderName;
  String folderName() {return folderName;}
 
  /* The number of messages in the currently open folder
   */
  private int numberOfMessages;
  int numberOfMessages() {return numberOfMessages;}
 
  /* The number of unread messages in the currently open folder
   */
  private int numberOfUnreadMessages;
  int numberOfUnreadMessages() {return numberOfUnreadMessages;}
  
  /* An array containing the messageIDs of messages in the currently 
   * open folder. messageIDs are assigned in the order messages are  
   * received from the server.  ?? Do we need this ??
   */
  private int[] messageIDArray;
  int messageID(int x) {return messageIDArray[x];}
  
  /* An array containing the exact sizes in octets of message in the
   * currently open folder.  Ordered by messageID.
   */
  private int[] messageSizeArray;
  int messageSize(int x) {return messageSizeArray[x];}

  /* An array containing the subjects of the messages in the currently 
   * open folder.  Ordered by messageID.
   */
  private String[] subjectArray;
  String subject(int x) {return subjectArray[x];}
 
  /* An array containing the from addresses of the messages in the currently 
   * open folder.  Ordered by messageID.
   */
  private String[] fromArray;
  String from(int x) {return fromArray[x];}

  /* An array containing the to addresses of the messages in the currently 
   * open folder.  Ordered by messageID.
   */
  private String[] toArray;
  String to(int x) {return toArray[x];} 

  /* An array containing the dates of the messages in the currently 
   * open folder.  Ordered by messageID.
   */
  private String[] dateArray;
  String date(int x) {return dateArray[x];} 


  /* An array of type boolean indicating whether or not each message
   * in the currently open folder has been read. Ordered by messageID.
   */ 
  private boolean[] readArray;
  boolean read(int x) {return readArray[x];}
  
/**
 * The following variables and methods deal with the current  
 * message.  The current message is selected by calling
 * the method getMessage(String messageID).
 */
 
  /* The messageID of the current message.
   */ 
  private int messageID;
  int messageID() {return messageID;}

  /* The date of the current message.
   */ 
  private String date;
  String date() {return date;}

  /* The subject field of the current message
   */
  private String subject;
  String subject() {return subject;}

  /* The to field of the current message
   */
  private String to;
  String to() {return to;}

  /* The from field of the current message
   */
  private String from;
  String from() {return from;}

  /* The reply-to field of the current message
   */
  private String replyTo;
  String replyTo() {return replyTo;}
  
  /* The cc field of the current message
   */
  private String cc;
  String cc() {return cc;} 

  /* The bcc field of the current message
   */
  private String bcc;
  String bcc() {return bcc;} 

  /* The full header of the current message
   */
  private String header;
  String header() {return header;}

  /* The body of the current message (unencrypted if the message was secure)
   */
  private String body;  
  String body() {return body;}

  /* The body of the current message (encrypted if the message was secure)
   */
  private String actualbody;
  String actualbody() {return actualbody;}

  /*  If true the message has been rewritten
   *  to the server, so the folderlist needs to be reloaded
   */
  private boolean wasRewritten;
  boolean wasRewritten() {return wasRewritten;}


  /*  This method wipes the entire message from memory
   */
  void wipeMessage()
  {
    date = null;
    subject = null;
    to = null;
    from = null;
    replyTo = null;
    cc = null;
    bcc = null;
    header = null;
    actualbody = null;
    body = "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------"+
           "-----------------------------------------------------------";
    body = null;
  }

/**
 * The following variables and methods deal with the addressbook  
 * message.  The addressbook is obtained by calling the method
 * getAddresses().
 */

  /* The number of addresses in the address book
   */
  private int numberOfAddresses;
  int numberOfAddresses() {return numberOfAddresses;}

  /* An array of nicknames for the entries in the addressbook  
   */
  private String[] nicknameArray;
  String nickname(int x) {return nicknameArray[x];}

  /* An array of fullnames for the entries in the addressbook
   */
  private String[] fullnameArray;
  String fullname(int x) {return fullnameArray[x];}  

  /* An array of email addresses for the entries in the addressbook
   */
  private String[] addressesArray;
  String addresses(int x) {return addressesArray[x];}

  /* An array of public key fingerprints
   */
  private String[] fingerprintsArray;
  String fingerprints(int x) {return fingerprintsArray[x];}

/**
 * The following variables and methods deal with information  
 * about the users account and current session.
 */

  /* The server's address
   */
  private String serverAddress;

  /* The port the server listens for connections on
   */
  private int serverPort = 21;

  /* The socket used to communicate with the server
   */
  private Socket s;  
 
  /* The service's domain.  Usually hushmail.com
   */
  private String domain;
  String domain() {return domain;}

  /* The username
   */
  private String username; 
    String username() {return username;}

  /* The users email address
   */
  private String emailAddress;
  String emailAddress() {return emailAddress;}  

  /* Half of a hash of the passPhrase;
   */
  private byte[] halfPassphraseHash;

  void setPassphrase(String passedPassphrase) 
  {
    byte[] passphraseHash = new HushSHA1().SHA1Hash(passedPassphrase);
    halfPassphraseHash=new byte[passphraseHash.length/2];
    System.arraycopy(passphraseHash,0,halfPassphraseHash,0,halfPassphraseHash.length);
    passphraseHash = null;
    passCipher = new BlowfishCipher();
    passCipher.setKey(passedPassphrase.getBytes());
    passedPassphrase = "                                            " +
                       "                                            " +
                       "                                            " +
                       "                                            " +
                       "                                            " ;
    passedPassphrase=null;
  }

  /*  The user's public key fingerprint
   */
  private String fingerprint = "";
  String fingerprint() {return fingerprint;}

  /* The cipher with the key as user's passphrase
   */
  private BlowfishCipher passCipher;
  void killPassCipher() {passCipher=null;}

  /* The pseudo stream cipher for communication between applet and server
   */
  private BlowfishCipher blowfishPipe;
 
  /* The sessionID.  64 bytes used as a unique identifier by the server to ensure
   * a user cannot open multiple sessions, and that no one can break into the middle
   * of a session.
   */
  private byte[] sessionIDBytes;

  /* The user's encrypted public key
   */
  private byte[] encPrivKey;

  /* A random number stream for encryption
   */
  private HushRandom randStream; 

  /* The applet using this instance of Connection
   */
  private HushApplet hushApplet; 

  /* A count of the tasks currently requiring use of the socket.
   * Allows the socket to close when not needed.
   */
  private int socketRequests;

  /* An output stream to the socket
   */
  BufferedOutputStream out;

  /* An input stream from the socket.
   */
  BufferedInputStream eIn;

  /* An input stream from a decrypted String which is used as a buffer
   * to store data from eIn after it has been decrypted with blowfishPipe
   */
  DataInputStream in;

  /* A boolean flag that indicates if the socket connection has been used yet.
   * This ensures the session will only be initialized by method use() once
   */
  private boolean initialUse;

  private String spaces="                               ";

/************************************************************
 *  Actions
 ************************************************************/
  
  Connection(String passedUsername, String passedSessionID, String passedSessionKey, HushApplet a) //Constructor
  {
    /**
     * This is the class constructor
     */
    username = passedUsername.toLowerCase();

    /* Convert sessionID hex string to an array of bytes
     */
    sessionIDBytes = Conversions.hexStringToBytes(passedSessionID);  

    /* Convert sessionKey hex string to an array of bytes
     */
    byte[] sessionKeyBytes = Conversions.hexStringToBytes(passedSessionKey);

    /* Instantiate cipher with sessionKey
     */
    blowfishPipe = new BlowfishCipher();
    blowfishPipe.setKey(sessionKeyBytes);

    hushApplet = a;
    serverAddress = hushApplet.getCodeBase().getHost();
    domain = "hushmail.com";
    emailAddress = username+"@"+domain;

    /* Since this session has just been started, no tasks are using the socket yet
     */
    socketRequests = 0;

    /* The socket has not yet been used for communication
     */
    initialUse = true;

    randStream = new HushRandom(new HushSHA1().SHA1Hash(Conversions.longToBytes(new Date().getTime())));

  }  


  void getFolders() throws IOException
  {  
   /**
    *  Retrieves information on the folders stored on the server
    *       numberOfFolders is set to the number of folders the 
    *           user has.
    *       folderNameArray contains the names of the folders.
    *        numberOfMessagesArray contains the numbers of messages
    *          in each folder 
    *        numberOfUnreadMessagesArray contains the numbers of 
    *          unread messages in each folder 
    */
    
    encryptAndWrite("LISTFOLDERS\n");
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();
    numberOfFolders = Integer.parseInt(in.readLine());
    folderNameArray = new String[numberOfFolders];
    numberOfMessagesArray = new int[numberOfFolders];
    numberOfUnreadMessagesArray = new int[numberOfFolders];
    for (int x=0;x<numberOfFolders;x++)     
    {
      folderNameArray[x] = in.readLine();
      numberOfMessagesArray[x] = Integer.parseInt(in.readLine()); 
      numberOfUnreadMessagesArray[x]= Integer.parseInt(in.readLine());
    }
  }

  void deleteFolder(String passedFolderName) throws IOException
  {
   /**
    *  Indicates to the server that the folder with name
    *  passedFolderName should be deleted
    */ 

    encryptAndWrite("DELETEFOLDER\n"+passedFolderName.trim()+"\n");
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();
   } 

  void newFolder(String newName) throws IOException
  {
   /**
    *  Indicates to the server that a new  folder with name
    *  newName should be created
    */ 
    encryptAndWrite("NEWFOLDER\n"+newName.trim()+"\n");
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();
  } 

    
  void renameFolder(String passedFolderName, String newName) throws IOException
  {
   /**
    *  Indicates to the server that the folder indicated by
    *  passedFolderName should be renamed newName
    */
    encryptAndWrite("RENAMEFOLDER\n"+passedFolderName.trim()+"\n"+newName.trim()+"\n");
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();
    if (folderName.equals(passedFolderName)) folderName = newName;
  }


  void openFolder(String passedFolderName) throws IOException
  {
   /**
    *   Retrieves information on the messages in a folder from the server
    *       folderName is set to passedFolderName.
    *        Currently open folder is folderName.
    *        numberOfMessages is set to the number of messages in  
    *          the currently open folder.
    *        numberOfUnreadMessages is set to the number of unread 
    *          messages in the currently open folder.
    *        messageIDArray, subjectArray, fromArray, readArray are  
    *          set for the messages in the currently open folder.
    */
    
    encryptAndWrite("LISTMESSAGES\n"+passedFolderName.trim()+"\n");
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();
    folderName = passedFolderName.trim();
    numberOfMessages = Integer.parseInt(in.readLine());
    messageIDArray = new int[numberOfMessages];
    subjectArray = new String[numberOfMessages];
    fromArray = new String[numberOfMessages];
    toArray = new String[numberOfMessages];
    readArray = new boolean[numberOfMessages];
    dateArray = new String[numberOfMessages];
    messageSizeArray = new int[numberOfMessages];   

    for (int x=0;x<numberOfMessages;x++)
    { 
      messageIDArray[x] = Integer.parseInt(in.readLine());
      readArray[x] = (Integer.parseInt(in.readLine())==1);
      subjectArray[x] = in.readLine();
      fromArray[x] = in.readLine();
      toArray[x] = in.readLine()+"";
      dateArray[x] = in.readLine();
      messageSizeArray[x] = Integer.parseInt(in.readLine());
    }
  }


  void deleteMessage(int passedMessageID) throws IOException
  {
   /**
    *  Indicates to the server to remove a message from
    *  the currently open folder
    */
    encryptAndWrite("DELETEMESSAGE\n"+folderName+"\n"+String.valueOf(passedMessageID)+"\n");
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();
   }


  void moveMessage(int messageID, String passedFolderName) throws IOException
  {
   /**
    *      Indicates to the server to move a message from the  
    *      currently open folder to the folder indicated by
    *      passedFolderName 
    */
   
      encryptAndWrite("MOVEMESSAGE\n"+folderName +"\n"+String.valueOf(messageID)+
        "\n"+passedFolderName.trim()+"\n"); 
      readAndDecrypt();
      if (!in.readLine().equals("OK")) throw new IOException();
  }
  

  /**
   * Retrieves message from the server:
   *    messageID is set to passedMessageID.
   *        date, subject, from, replyTo, to, cc, body are set to 
   *          the corresponding fields from the message in the 
   *          currently open folder with messageID equal to
   *          passedMessageID. 
   *
   */
  void getMessage(int passedMessageID) throws IOException
  {
    /* Clear old message info
     */
   
    messageID = passedMessageID;
    date="";
    subject="";
    to=""; 
    from="";
    replyTo="";
    cc="";
    bcc="";

    /*  This field contains information on the message's encryption.  For example it may contain
     *  an encrypted key that is used to symmetrically encrypt the actual message, or it may
     *  indicated that the message body is symmetrically encrypted directly with the user's passphrase
     */ 
    String encryption="";
    String keyblock="";
    body="";
    actualbody="";
    wasRewritten = false;
  
    encryptAndWrite("GETMESSAGE\n"+folderName+"\n"+String.valueOf(passedMessageID)+"\n");
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();

    int lengthOfMessage = Integer.parseInt(in.readLine());

    /* Message must be converted to bytes to get around
     * problems with string reading in Java 1.0
     */
    byte[] messageBytes = new byte[lengthOfMessage];
    in.read(messageBytes);
    DataInputStream messageParser = new DataInputStream(new ByteArrayInputStream(messageBytes)); 

    String line; 
      
    //  Determine whether message is encrypted
    boolean encrypted = false; 

    String eTo = "";
    String eCc = "";
    String eBcc = "";

    String lastField = new String();
    while ((line=messageParser.readLine())!=null)
      if (line.length()>4) 
      {
        header=line.trim()+"\n";
        break;
      }

    /* This loop parses out header fields
     * and compensates for header modifications necessary
     * in encrypted messages
     */ 
    while ((line=messageParser.readLine())!=null)
    {
      if (line.length()==0) break;
      else 
      {
       /* A string representing the complete header
        * accessed if user wants to see full header
        */
        header=header+line+"\n";

       /* Concatenate information from wrapped header fields
        */

        if (line.regionMatches(0," ",0,1))
        { 
          if (lastField.equals("to")) to=to+line.trim();
          else if (lastField.equals("cc")) cc=cc+line.trim();
          else if (lastField.equals("bcc")) bcc=bcc+line.trim();
          else if (lastField.equals("eTo")) eTo=eTo+line.trim();
          else if (lastField.equals("eCc")) eCc=eCc+line.trim();
          else if (lastField.equals("eBcc")) eBcc=eBcc+line.trim();
          else if (lastField.equals("date")) date=date+line.trim();
          else if (lastField.equals("subject")) subject=subject+line.trim();
          else if (lastField.equals("replyTo")) replyTo=replyTo+line.trim();
          else if (lastField.equals("encryption")) encryption=encryption+line.trim();
          else if (lastField.equals("keyblock")) keyblock=keyblock+line.trim();
        }
        else if (line.regionMatches(0,"From:",0,5)) 
        {
          from=line.substring(5).trim();
          lastField = "from";
        }
        else if (line.regionMatches(0,"Date:",0,5)) 
        {
          date=line.substring(5).trim();
          lastField = "date";
        }
        else if (line.regionMatches(0,"To:",0,3)) 
        {
          to=line.substring(3).trim();
          lastField = "to";
        }
        else if (line.regionMatches(0,"Cc:",0,3)) 
        {
          cc=line.substring(3).trim();
          lastField = "cc";
	} 
        else if (line.regionMatches(0,"Bcc:",0,4)) 
        {
          bcc=line.substring(4).trim();
          lastField = "bcc";
	}
        /* E-cc is used to store Cc information in encrypted messages
         * which can have no Cc field 
         */ 
        else if (line.regionMatches(0,"E-Bcc:",0,6)) 
        {
          eBcc=line.substring(6).trim();
          lastField = "eBcc";
        } 
        /* E-to overrides To in encrypted messages 
         */
        else if (line.regionMatches(0,"E-To:",0,5)) 
        {
          eTo=eTo+line.substring(5).trim();
          lastField = "eTo";
        }

        /* E-cc is used to store Cc information in encrypted messages
         * which can have no Cc field 
         */
        else if (line.regionMatches(0,"E-Cc:",0,5)) 
        {
          eCc=line.substring(5).trim();
          lastField = "eCc";
        } 
        else if (line.regionMatches(0,"Subject:",0,8)) 
        {
          subject=line.substring(8).trim();
          lastField = "subject";
        } 
        else if (line.regionMatches(0,"Reply-To:",0,9)) 
        {
          replyTo=line.substring(9).trim();
          lastField = "replyTo";
        }  
        else if (line.regionMatches(0,"Hush-encryption:",0,16)) 
        {
          encrypted = true;
          encryption=line.substring(17).trim();
          lastField = "encryption";
        } 
        else if (line.regionMatches(0,"Hush-keyblock:",0,14))
        {
          keyblock = line.substring(15).trim();
          lastField = "keyblock";
        } 
        else lastField = "";   
      }
    }  // end while (line) loop


    if (encrypted) to="";
    if (eTo.length()>0) to=eTo;
    if (eCc.length()>0) cc=eCc;
    if (eBcc.length()>0) bcc=eBcc;

    StringBuffer bodyBuffer = new StringBuffer();

    while ((line=messageParser.readLine())!=null)
    {
      bodyBuffer.ensureCapacity(bodyBuffer.length()+line.length());
      bodyBuffer = bodyBuffer.append(line + "\n");
    }
    body = bodyBuffer.toString();      

    to=wrap(to,true,false);
    cc=wrap(cc,true,false);
    bcc=wrap(bcc,true,false);
    replyTo=wrap(replyTo,true,false);
    subject=wrap(subject,true,false);
    actualbody = body;

    if (encrypted)
    {
      if (!folderName().equals("drafts")) 
        hushApplet.statusBar(spaces+"This message was sent securely"+spaces);
 
      /* The body of the message should be decrypted here
       * using the variable 'body' and the information stored in
       * the variable 'encryption'
       */ 
      byte[] bodyBytes = new byte[body.length()];
      body.getBytes(0,body.length(),bodyBytes,0);
      DataInputStream parseBody = new DataInputStream(new ByteArrayInputStream(bodyBytes));
      StringBuffer encBodyBuf = new StringBuffer();
      String bodyHash = "";
      boolean inCipher = false;
      String i;
      while ((i = parseBody.readLine())!=null)
      {
        i = i.trim();
        if (i.equals("----- HushMail v1.0 -----")) inCipher=true;
        else if (i.regionMatches(0,"-",0,1))
        {
          bodyHash = i.substring(1);
          break;
        }
        else if (inCipher) 
        {
          encBodyBuf.ensureCapacity(encBodyBuf.length()+i.length());
          encBodyBuf.append(i);
        }
      }
      byte[] encKeyblock = Conversions.hexStringToBytes(keyblock);
      byte[] plainKey = new byte[16];         
      byte[] keyHash = new byte[20];
      byte[] bodyKeyAndHash = new byte[36];

       /* Decrypt bodykey based on information included in "Encryption:" header
        */
      if (encryption.equals("Hush Private 1.0")) bodyKeyAndHash = passCipher.decrypt(encKeyblock);
      else if (encryption.equals("Hush Public 1.0")) 
      {
        /*  Make sure everything is wiped after private key is decrypted and used
         */
        byte[] privKey = passCipher.decrypt(encPrivKey);
        ElGamalCipher privKeyCipher = new ElGamalCipher();
        privKeyCipher.setPrivateKey(privKey);
        bodyKeyAndHash = privKeyCipher.hushDecrypt(encKeyblock,36);
        for (int x=0;x<privKey.length;x++) privKey[x]=0;
        privKeyCipher.setPrivateKey(privKey);
        privKey=null;
        privKeyCipher=null;
      }

      System.arraycopy(bodyKeyAndHash,0,plainKey,0,16);
      System.arraycopy(bodyKeyAndHash,16,keyHash,0,20);

      if (keyHash.length!=20) hushApplet.statusBar(spaces+"Wrong hash length.  Message may not be valid."+spaces);
      else
      {
        if (!Conversions.bytesToHexString(keyHash).equals(Conversions.bytesToHexString(new HushSHA1().SHA1Hash(plainKey)))) hushApplet.statusBar(spaces+"Hash failed.  Message may not be valid"+spaces);
      }

      BlowfishCipher bodyCipher = new BlowfishCipher();
      bodyCipher.setKey(plainKey);
      body = bodyCipher.stringDecrypt(Conversions.hexStringToBytes(encBodyBuf.toString()));
      if (bodyHash.length()!=40) 
        hushApplet.statusBar(spaces+"Wrong hash length.  Message may not be valid."+spaces);
      {
        if (!bodyHash.equals(Conversions.bytesToHexString(new HushSHA1().SHA1Hash(body)))) 
          hushApplet.statusBar(spaces+"Hash failed.  Message may not be valid"+spaces);
      }
      if (encryption.equals("Hush Public 1.0"))
      {
        
        String newKeyblock = "Hush-keyblock: "+Conversions.bytesToHexString(passCipher.encrypt(bodyKeyAndHash))+"\n";
        StringBuffer newMessageBuf = new StringBuffer();
        boolean inKeyblock = false;
        boolean foundHeader = false;
        boolean statusFieldFound = false;
        messageParser.reset();
        while ((line=messageParser.readLine())!=null)
        {
          newMessageBuf.ensureCapacity(newMessageBuf.length()+line.length());
          if (line.regionMatches(0,"From ",0,5)) foundHeader=true;
          if (line.trim().length()==0 && !statusFieldFound && foundHeader)
	  {
            statusFieldFound = true;
            newMessageBuf.append("Status: RO\n");
          }
          if (line.regionMatches(0,"Hush-encryption:",0,16)) newMessageBuf.append("Hush-encryption: Hush Private 1.0\n");
          else if (line.regionMatches(0,"Hush-keyblock:",0,14))
          {
            inKeyblock = true;
            newMessageBuf.append(wrap(newKeyblock,false,true));
          }
          else if (line.regionMatches(0,"Status:",0,7))
	  {
            inKeyblock = false;
            newMessageBuf.append("Status: RO\n");
          }
          else if (!line.regionMatches(0," ",0,1)&&inKeyblock)
          {
            inKeyblock = false;
            newMessageBuf.append(line + "\n");
          }
          else if (!inKeyblock) newMessageBuf.append(line + "\n"); 
        }
        deleteMessage(messageID);
        String newMessage = newMessageBuf.toString();
        encryptAndWrite("SAVEMESSAGE\n" + folderName + "\n" + String.valueOf(newMessage.length()) + "\n" + newMessage);
        readAndDecrypt();
        if (!in.readLine().equals("OK")) throw new IOException();
        wasRewritten = true;
      } 
    }
    else 
    {
      if (!folderName().equals("drafts")) 
        hushApplet.statusBar(spaces+"This message was NOT sent securely"+spaces);
      else hushApplet.statusBar(spaces+"This message was NOT saved securely"+spaces);
    }
    body = wrap(body,true,false);
  }


  void sendPlainMessage(String header, String body, boolean saveToSent)
    throws IOException
  {
    String message = new String(header+"\n"+body);
    if (saveToSent) 
    {
      String firstLine =  "From "+emailAddress()+"  "+new Date().toString()+"\n";
      encryptAndWrite("SAVEMESSAGE\nsent\n"+String.valueOf(firstLine.length()+message.length())+"\n"+firstLine+message);
      readAndDecrypt();
      if (!in.readLine().equals("OK")) throw new IOException();
    }
    encryptAndWrite("SENDMESSAGE\n"+String.valueOf(message.length())+"\n"+message);
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();
  }


  /**
   *  Send a formatted message to the server.
   *   NOTE!!  Line wraps in header will be done on server side!!
   */
  boolean sendEncryptedMessage(String[] recipients, String[] headers, String body, boolean save, String saveFolder) throws IOException
  {
    /* Returns true if all recipients are vaild HushMail users.
     */

    /* Get the public keys
     */
      
    String[] pubKeys = new String[recipients.length];
   
    for (int x=0;x<recipients.length;x++)
    {
      if (!recipients[x].equals(username)&&!recipients[x].equals("/self/"))
      {
        hushApplet.statusBar(spaces+"Retrieving public key for "+recipients[x]+spaces);
        encryptAndWrite("GETPUBLICKEY\n"+recipients[x]+"\n");
        readAndDecrypt();
        if (!in.readLine().equals("OK")) throw new IOException();        

        if (in.readLine().equals("APPROVED")) pubKeys[x] = in.readLine().trim();
        else 
        {
          hushApplet.statusBar (spaces+"No such user: "+recipients[x]+spaces);
          return false;
        }

        for (int y=0;y<numberOfAddresses;y++)
	{
          
          if (addressesArray[y].indexOf("@"+domain)==addressesArray[y].length()-("@"+domain).length() && recipients[x].regionMatches(0,addressesArray[y],0,addressesArray[y].indexOf("@")))
	  {
            String fingerprintReceived = Conversions.bytesToHexString(new HushSHA1().SHA1Hash(Conversions.hexStringToBytes(pubKeys[x])));

            if (!fingerprints(y).equals(fingerprintReceived))
	    {
              hushApplet.statusBar(spaces+"Bad public key for "+recipients[x]+spaces);
              return false;
            }
          }
        }
      }
    }

     /*  Encrypt the body with a randomly generated 
      *  Blowfish key and append SHA1 Hash
      */

    byte[] bodyKey = new byte[16];
    randStream.nextBytes(bodyKey);
    BlowfishCipher bodyCipher = new BlowfishCipher();
    bodyCipher.setKey(bodyKey);
    hushApplet.statusBar(spaces+"Encrypting message body"+spaces);
    body = wrap("----- HushMail v1.0 -----\n"+
      Conversions.bytesToHexString(bodyCipher.stringEncrypt(body)) + 
      "\n-" + Conversions.bytesToHexString(new HushSHA1().SHA1Hash(body)) + 
      "\n----- End -----\n",false,false);

    hushApplet.statusBar(spaces+"Body encrypted"+spaces);

    /*  Concatenate the body key and a SHA1Hash of the body key into one array
     */ 
    
    byte[] bodyHash = new HushSHA1().SHA1Hash(bodyKey);
    byte[] bodyKeyAndHash = new byte[bodyKey.length+bodyHash.length];
    System.arraycopy(bodyKey,0,bodyKeyAndHash,0,bodyKey.length);
    System.arraycopy(bodyHash,0,bodyKeyAndHash,bodyKey.length,bodyHash.length);

    for (int x=0;x<recipients.length;x++)
    {
      String encryption = "Hush-encryption: ";

      if (recipients[x].equals(username)||recipients[x].equals("/self/"))
      { 
          /*  If recipient is self
	   *  Blowfish the body key  using the users passphrase
           */
          encryption = encryption + "Hush Private 1.0\n";
          encryption = encryption + wrap("Hush-keyblock: " + 
            Conversions.bytesToHexString(passCipher.encrypt(bodyKeyAndHash)), false, true) + "\n";
      } 
      else
      {
        hushApplet.statusBar(spaces+"Encrypting to "+recipients[x]+spaces);
        ElGamalCipher pubKeyCipher = new ElGamalCipher();
        pubKeyCipher.setRandomStream(randStream);
        pubKeyCipher.setPublicKey(Conversions.hexStringToBytes(pubKeys[x]));
        encryption = encryption + "Hush Public 1.0\n";
        encryption = encryption + wrap("Hush-keyblock: " + 
          Conversions.bytesToHexString(pubKeyCipher.hushEncrypt(bodyKeyAndHash,bodyKeyAndHash.length)), 
          false, true) + "\n";
      }

      String message = new String(headers[x]+encryption+"\n"+body);
      
      if (save || recipients[x].equals("/self/"))
      {
        hushApplet.statusBar(spaces+"Saving copy in '"+saveFolder+"'"+spaces);
        encryptAndWrite("SAVEMESSAGE\n"+saveFolder+"\n"+String.valueOf(message.length())+"\n"+message);
        readAndDecrypt();
        if (!in.readLine().equals("OK")) throw new IOException();       
      }
      else
      {
        hushApplet.statusBar(spaces+"Sending message"+spaces);
        encryptAndWrite("SENDMESSAGE\n"+String.valueOf(message.length())+"\n"+message);
        readAndDecrypt();
        if (!in.readLine().equals("OK")) throw new IOException();       
      }
    }
    return true;
  }


  /**
   *  Retrieves address information from server:
   *    numberOfAddresses set to number of addresses in 
   *         addressbook.
   *    The variables nicknameArray, fullnameArray, 
   *       and addressesArray, are set to appropriate values
   *           from addressbook on server
   */
  void getAddresses() throws IOException
  {
    encryptAndWrite("GETADDRESSES\n");
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();       
    numberOfAddresses = Integer.parseInt(in.readLine());
    nicknameArray = new String[numberOfAddresses];
    fullnameArray = new String[numberOfAddresses];
    addressesArray = new String[numberOfAddresses];
    fingerprintsArray = new String[numberOfAddresses];
    for (int x=0;x<numberOfAddresses;x++)
    { 
      nicknameArray[x] = in.readLine();
      fullnameArray[x] = in.readLine();
      addressesArray[x] = in.readLine();
      fingerprintsArray[x] = in.readLine();
    }
  }


  void deleteAddress(String passedNickname) throws IOException
  {
    encryptAndWrite("DELETEADDRESS\n"+passedNickname+"\n");
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();       
  }


  boolean addAddress(String passedNickname,String passedFullname,String passedAddress) throws IOException
  {
    passedFullname = passedFullname.replace('"','\'');
    String fingerprint = "";
    if ((passedAddress+"\n").indexOf("@"+domain+"\n")!=-1)
    {
      String newAddressUser = passedAddress.substring(0,passedAddress.indexOf("@"));
      encryptAndWrite("GETPUBLICKEY\n"+newAddressUser);
      readAndDecrypt();
      if (!in.readLine().equals("OK")) throw new IOException();       
      String response=in.readLine();
      if (response.equals("NOT APPROVED")) 
      {
        hushApplet.statusBar(spaces+"No such HushMail user: "+newAddressUser+spaces);
        return false;
      }
      else fingerprint=Conversions.bytesToHexString(new HushSHA1().SHA1Hash(Conversions.hexStringToBytes(in.readLine())));
    }
    encryptAndWrite("NEWADDRESS\n"+passedNickname+"\n"+passedFullname+"\n"+passedAddress+"\n"+fingerprint+"\n");    
    readAndDecrypt();
    if (!in.readLine().equals("OK")) throw new IOException();       
    return true;
  }


  String linewrap(String inString, int linelength)
  {
    StringBuffer outStringBuf = new StringBuffer();
    outStringBuf.ensureCapacity(inString.length()+inString.length()/linelength+2);
    int place = 0;
    for (int x=0;x<inString.length();x++)
    {
      outStringBuf.append(inString.substring(x,x+1));
      if (inString.substring(x,x+1).equals("\n")) place=0;
      else if (++place==linelength)
      {
        outStringBuf.append(inString.substring(x,x+1)+"\n");
        place=0; 
      }
    }
    return outStringBuf.toString();
  }


  String wordwrap(String inString, int linelength)
  {
    StringBuffer outStringBuf = new StringBuffer();
    outStringBuf.ensureCapacity(inString.length()+inString.length()/linelength+2);
    int place = 0;
    String line="";
    for (int x=0;x<inString.length();x++)
    {
      line = line + inString.substring(x,x+1);
      place++;
      if (line.regionMatches(line.length()-1,"\n",0,1)) place=0;
      else if (place>=linelength)
      { 
        int lastSpace = line.lastIndexOf(' ');
        if (lastSpace!=-1 && lastSpace!=0)
        {
          outStringBuf.append(line.substring(0,lastSpace)+"\n");
          if (lastSpace==line.length()-1) line=" ";
          else line=line.substring(lastSpace+1);
          place=line.length();
        }
      }
    }
    return outStringBuf.toString();
  }

 
  String wrap(String inString, boolean isParagraph, boolean isHeader)
  {
   /**
    *  This function wraps Strings longer then 72 characters
    */

    String space = "";
    int resetPlace = 0;
    if (isHeader) 
    {
      space = " ";
      resetPlace = 0;
    }
    String term = "\n";
    StringBuffer outStringBuf = new StringBuffer();
    outStringBuf.ensureCapacity(inString.length()*2);
    int place = 0;
    for (int x=0;x<inString.length();x++)
    {
      term = "\n";
      if (x+1<inString.length() && inString.charAt(x+1)=='\n') term = "";

      if (inString.charAt(x)=='\n') 
      {
        outStringBuf.append("\n");
        place=0;
      }
      else if (place==71 && !isParagraph)
      {
        outStringBuf.append(inString.charAt(x)+term+space);
        place=resetPlace; 
      }
      else if (place>=71 &&
(inString.charAt(x)==','||inString.charAt(x)=='-'|| inString.charAt(x)==' '))
      {
        outStringBuf.append(inString.charAt(x)+term+space);
        place=resetPlace;
      }
      else 
      {
        outStringBuf.append(inString.charAt(x));
        place++;
      }
    }
    return outStringBuf.toString();
  }  
  

  /**
   *  This method should be called when a task requires 
   *  the use of the socket connection 
   */
  boolean use() throws IOException
  {
    hushApplet.startBlinkingText();

    if (initialUse)
    {
      /* Increment count of threads using socket
       */
      ++socketRequests;
 
      s = new Socket(serverAddress,serverPort);
      s.setSoTimeout(60000);
      s.setTcpNoDelay(false); 

      out = new BufferedOutputStream(s.getOutputStream());
      eIn = new BufferedInputStream(s.getInputStream());

      /* Send sessionID to server, so that the server can find the sessionKey
       */
      out.write(sessionIDBytes);

      /* Send username and half of hash of passphrase
       */
      encryptAndWrite("CONNECT\n"+username+"\n"+Conversions.bytesToHexString(halfPassphraseHash)+"\n");

      /* Check for acceptance of username and passphrase
       */
      readAndDecrypt();
      if (!in.readLine().equals("OK")) throw new IOException();        
      if (!in.readLine().equals("APPROVED"))
      {
        /*  Exit to "BadLogin" page
         */
        try 
        {
          URL u = new URL(hushApplet.exitpage+"?action=BadLogin&username="+username());
          hushApplet.getAppletContext().showDocument(u);
        }  
        catch (MalformedURLException m) {System.out.println("MalformedURL");}    
        eIn.close();
        out.close();
        s.close();
        return false;
      }
      String publicKey = in.readLine();
      if (publicKey!=null) fingerprint = Conversions.bytesToHexString(new HushSHA1().SHA1Hash(Conversions.hexStringToBytes(publicKey)));
      String encPrivKeyString = in.readLine();      
      if (encPrivKeyString!=null) encPrivKey = Conversions.hexStringToBytes(encPrivKeyString);
      initialUse = false;
    }
    else
    {
      /* Execute only if no other threads are currently using the socket
       */
      if (++socketRequests==1)
      {
        s = new Socket(serverAddress,serverPort);
        s.setSoTimeout(180000);
        s.setTcpNoDelay(true);
        out = new BufferedOutputStream(s.getOutputStream());
        eIn = new BufferedInputStream(s.getInputStream());

        /* Send sessionID to server, so that the server can find the sessionKey
         */
        out.write(sessionIDBytes);

        encryptAndWrite("RECONNECT\n"+username+"\n");

        /* Check for approval for continuation of session
         */
        readAndDecrypt();
        if (!in.readLine().equals("OK")) throw new IOException();       
        if (!in.readLine().equals("APPROVED"))
        {
          try
          {
            URL u = new URL(hushApplet.exitpage+"?action=NotApproved&username="+username());
            hushApplet.getAppletContext().showDocument(u);
            hushApplet.removeAll();
            hushApplet.paintAll(hushApplet.getGraphics());
            wipeMessage();
            killPassCipher();
          }  
          catch (MalformedURLException m) {System.out.println("MalformedURL");} 
          eIn.close();
          out.close();
          s.close();
          return false;
        }
      }
    }
    return true;
  }


  /**
   *  This method should be called when a task using
   *  the socket connection has been completed
   */
  void done()
  {
    hushApplet.stopBlinkingText();

    if (s==null) return;

    /* decrement count of threads using socket
     * if count is zero then terminate the socket
     */
    if (--socketRequests==0)
    {
      try {
        out.close();
        eIn.close();
        s.close();
      } catch (IOException e1) {System.out.println("IOException");}
    }
  }


  /*  This signals the server to delete the session record
   *  and terminates the socket
   */
  void quit()
  {
    if (s==null) return;
    try
    {
      encryptAndWrite("QUIT\n");
      out.close();
      eIn.close();
      s.close();
    } catch (IOException e1) {System.out.println("IOException");}
  }


  /**
   *  This methods send the parameter string to the
   *  server, calling appropriate encryption methods 
   *  and indicating the length.
   */
  private void encryptAndWrite(String output)
  {
    try
    {
    /* Send 16 bytes indicating the length of the output to follow
     */
      byte[] encOutputBytes = blowfishPipe.stringEncrypt(output); 
      int outLength = encOutputBytes.length;
      out.write(blowfishPipe.encrypt(Conversions.intToBytes(outLength)));
      out.write(encOutputBytes);
      out.flush();	
    } catch (IOException e) {System.out.println("IOException");}
  }


  /**
   *  This method reads in a set of bytes from the server
   *  calls appropriate decryption methods, and stores a resulting
   *  string in a buffer for later access.
   */
  private void readAndDecrypt() throws IOException
  {
    /* read in 16 bytes indicating the length of the input to follow 
     */
    byte lengthBytes[] = new byte[16];

    int bytesRead = eIn.read(lengthBytes);
    while (bytesRead<16) bytesRead=bytesRead+eIn.read(lengthBytes,bytesRead,16-bytesRead);
      
    int inLength = Conversions.bytesToInt(blowfishPipe.decrypt(lengthBytes));

    if (inLength==0) throw new IOException();

    /* read and decrypt the requested number of bytes
     */
    byte[] eBuf = new byte[inLength];

    bytesRead = eIn.read(eBuf);
    while (bytesRead<inLength) bytesRead=bytesRead+eIn.read(eBuf,bytesRead,inLength-bytesRead);  

    String inString = blowfishPipe.stringDecrypt(eBuf);

    byte[] trimmedBytes = new byte[inString.length()];
    inString.getBytes(0,inString.length(),trimmedBytes,0);
    in = new DataInputStream(new ByteArrayInputStream(trimmedBytes));
  }

 
  /**
   * Variables and methods for a queue of bits
   * intended to store random values for later access
   */
  
  ByteQueue randBytes = new ByteQueue(); 

  void enqueueBit(long l, int bits)
  {
    randBytes.enqueueBits(l,bits);
    if (randBytes.bytesAvailable()>0)
    {
      byte[] available = randBytes.dequeueBytes(randBytes.bytesAvailable());
      randStream.enqueueBytes(available);
    }
  }
}   //  end Connection.java

